from typing import Literal as _Literal


Action = _Literal[
    "auth",
    "pay",
    "hold",
    "payqr",
    "paytoken",
    "paycash",
    "hold_completion",
    "subscribe",
    "paydonate",
    "paysplit",
    "split_rules",
    "letter_of_credit",
    "apay",
    "gpay",
    "regular",
    "refund",
    "payment_prepare",
    "p2pcredit",
    "p2pdebit",
    "confirm",
    "mpi",
    "cardverification",
    "register",
    "data",
    "ticket",
    "status",
    "invoice_send",
    "invoice_cancel",
    "token_create",
    "token_create_unique",
    "token_update",
    "reports",
    "reports_compensation",
    "reports_compensation_file",
    "agent_shop_create",
    "agent_shop_edit",
    "agent_info_merchant",
    "agent_info_user",
    "unsubscribe",
    "subscribe_update",
]
